/**
 * @file gCardSys_Controller.cc
 *
 *  @remark ~ Implementation of the LeinerSys Controller module
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-03 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>              // needed by tolower

#include <unistd.h>             // need by get_current_dir_name
#include <gdk/gdkkeysyms.h>

#ifdef IS_HILDON
#include <hildon/hildon-program.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <esd.h>
#endif


#include "gCardSys_Controller.hh"

/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;
/* TC 2010-05-09 (Sun) -- copied from osso-af-utils: */
/* The full volume should be about 30% of maximum, that is 0x4C out of 0xFF.
   The full volume is marked as 2 in gconf, divide and we get 0x26. */
// #define SCALE_MULTIPLIER 0x26   // for maemo


/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[gCardSys_Controller %d]: ",       \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam(...)
#else
static void fakeUseParam(...) {}
#endif


#ifdef IS_HILDON
inline static GtkWidget* 
hildon_builder_get_object (GtkBuilder *builderP,
                           const char *name_str)
{
  GtkWidget *content_vboxP =
    GTK_WIDGET (gtk_builder_get_object (builderP, "content_vbox"));
  assert (content_vboxP);

  /* replacement widgets: */
  GtkWidget *widgetP =  
    GTK_WIDGET (g_object_get_data (G_OBJECT (content_vboxP), name_str));

  if (unlikely (widgetP == NULL))
    {
      /* first time getting main_window: */
      if (unlikely (strcmp (name_str, "main_window") == 0))
        {
          widgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                        "main_window"));
          g_object_set_data (G_OBJECT (content_vboxP), 
                             "content_vbox", content_vboxP);

          fprintf (stderr, "looking for main_window %p from builder\n",
                   widgetP);
        }
      widgetP = GTK_WIDGET (gtk_builder_get_object (builderP, name_str));
    }

  return widgetP;
}

#define gtk_builder_get_object hildon_builder_get_object
#endif


/*******************************************\
 * Callback functions and helpers:
\*******************************************/

// #include <sys/time.h>           // needed for gettimeofday
// static double 
// mstime (void)
// {
//   double m;
//   struct timeval tp;
//   if (gettimeofday(&tp, 0) != 0) // obsolete ...  TBF
//     return 0;
//   m = ((double) tp.tv_usec / 1000.0) + (tp.tv_sec * 1000);
//   return m;
// }



// extern "C" G_MODULE_EXPORT  void 
// on_main_window_destroy_event (GtkWidget *widget,
//                               gpointer   user_data)

extern "C" G_MODULE_EXPORT  gboolean
on_main_window_destroy_event (GtkWidget *widget,
                              GdkEvent  *event,
                              gpointer   user_data)
{
  fakeUseParam (widget, event, user_data);
  fprintf (stderr, "hello2\n");
  gtk_main_quit ();  
  return TRUE;                  // handled.
}


extern "C" G_MODULE_EXPORT void  
on_handheld_chkmItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                              gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget  *gtkWidgetP = NULL;

  /*
   * Switch to handheld mode:
   */
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    {
      fprintf (stderr, "handheld check item checked\n");

      /* hide prev_btn and next_btn: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "prev_btn"));
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);

      /* show nopeNext_btn and yepNext_btn and enable if is opened: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                       "nopeNext_btn"));
      assert (gtkWidgetP);
      if (ctrlObjP->m_isOpened_flg)
        gtk_widget_set_sensitive (gtkWidgetP, TRUE);
      gtk_widget_show (gtkWidgetP);
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                       "yepNext_btn"));
      assert (gtkWidgetP);
      if (ctrlObjP->m_isOpened_flg)
        gtk_widget_set_sensitive (gtkWidgetP, TRUE);
      gtk_widget_show (gtkWidgetP);


      /* hide yep_radio and nope_radio: */
#ifdef IS_HILDON
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "mark_btn"));
#else
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "mark_vbox"));
#endif
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);


      /* show the handheld back button: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object 
                               (builderP, "handHeldBack_btn"));
      assert (gtkWidgetP);
      gtk_widget_show (gtkWidgetP);
    }
  else

  /*
   * Switch to non-handheld mode:
   */
    {
      /* show prev_btn and next_btn: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "prev_btn"));
      assert (gtkWidgetP);
      gtk_widget_show (gtkWidgetP);
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
      assert (gtkWidgetP);
      gtk_widget_show (gtkWidgetP);

      /* hide nopeNext_btn and yepNext_btn: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                       "nopeNext_btn"));
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                       "yepNext_btn"));
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);

      /* enables yep_radio and nope_radio if is opened*/
      if (ctrlObjP->m_isOpened_flg)
        {
          gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                           "mark_vbox"));
          assert (gtkWidgetP);
          gtk_widget_set_sensitive (gtkWidgetP, TRUE);
        }

      /* show yep_radio and nope_radio: */
#ifdef IS_HILDON
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "mark_btn"));
#else
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "mark_vbox"));
#endif
      assert (gtkWidgetP);
      gtk_widget_show (gtkWidgetP);

      /* hide the handheld back button: */
      gtkWidgetP = GTK_WIDGET (gtk_builder_get_object 
                               (builderP, "handHeldBack_btn"));
      assert (gtkWidgetP);
      gtk_widget_hide (gtkWidgetP);
    }
}


extern "C" G_MODULE_EXPORT void  
on_flipToFront_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                   gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_viewObjP->m_attributeObj.SetValue 
      ("autoFlip_enum", AUTOFLIP_TO_FRONT);

  /*
   * switch to Front now:
   */
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkNotebook *the_tabsP = 
    GTK_NOTEBOOK (gtk_builder_get_object (builderP, "the_tabs"));
  gint curTag_int = gtk_notebook_get_current_page (the_tabsP);
  if (curTag_int != ctrlObjP->m_frontPageNum_int)
    gtk_notebook_set_current_page (the_tabsP, ctrlObjP->m_frontPageNum_int);
}


extern "C" G_MODULE_EXPORT void  
on_flipToBack_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                  gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_viewObjP->m_attributeObj.SetValue 
      ("autoFlip_enum", AUTOFLIP_TO_BACK);

  /*
   * switch to Back now:
   */
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkNotebook *the_tabsP = 
    GTK_NOTEBOOK (gtk_builder_get_object (builderP, "the_tabs"));
  gint curTag_int = gtk_notebook_get_current_page (the_tabsP);
  if (curTag_int != ctrlObjP->m_backPageNum_int)
    gtk_notebook_set_current_page (the_tabsP, ctrlObjP->m_backPageNum_int);
}


extern "C" G_MODULE_EXPORT void  
on_flipOff_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                               gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_viewObjP->m_attributeObj.SetValue 
      ("autoFlip_enum", AUTOFLIP_OFF);
}


extern "C" G_MODULE_EXPORT void  
on_soundOnFront_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                    gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_attributeObjP->SetValue 
      ("autoSound_enum", AUTOSOUND_ON_FRONT);
}

extern "C" G_MODULE_EXPORT void  
on_soundOnBack_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                    gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_attributeObjP->SetValue 
      ("autoSound_enum", AUTOSOUND_ON_BACK);
}

extern "C" G_MODULE_EXPORT void  
on_soundOff_radiomItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                gpointer          in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    ctrlObjP->m_attributeObjP->SetValue 
      ("autoSound_enum", AUTOSOUND_OFF);
}


extern "C" G_MODULE_EXPORT void  
on_reviewLastBoxOnly_chkmItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                       gpointer          in_userData)
{
  // fprintf (stderr, "Calling reviewLastBoxOnly\n");
  
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget * gtkWidgetP = NULL;

  /* toggle the setting: */
  if (gtk_check_menu_item_get_active (in_checkMenuItemP))
    {
      /* disable review2ndBoxOnly_chkmItem first: */
      gtkWidgetP = 
        GTK_WIDGET (gtk_builder_get_object (builderP, 
                                            "review2ndBoxOnly_chkmItem"));
      assert (gtkWidgetP);
      gtk_check_menu_item_set_active // will call the callback function
        (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);

      /* set the cardbox: */
      ctrlObjP->m_modelObjP->SetReviewLastBoxOnly (true);
    }
  else
    ctrlObjP->m_modelObjP->SetReviewLastBoxOnly (false);
}


extern "C" G_MODULE_EXPORT void  
on_review2ndBoxOnly_chkmItem_toggled (GtkCheckMenuItem *in_checkMenuItemP,
                                      gpointer          in_userData)
{
  // fprintf (stderr, "Calling review2ndBoxOnly\n");
  
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget *gtkWidgetP = NULL;
  

  /* toggle the setting: */
  if (gtk_check_menu_item_get_active  (in_checkMenuItemP))
    {
      /* disable reviewLastBoxOnly_chkmItem first: */
      gtkWidgetP = 
        GTK_WIDGET (gtk_builder_get_object (builderP, 
                                            "reviewLastBoxOnly_chkmItem"));
      assert (gtkWidgetP);
      gtk_check_menu_item_set_active // will call the callback function
        (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);

      /* set the cardbox: */
      ctrlObjP->m_modelObjP->SetReview2ndBoxOnly (true);
    }
  else
    ctrlObjP->m_modelObjP->SetReview2ndBoxOnly (false);
}


static void
PrintCurStudyCard (ControllerClass *ctrlObjP)
{
  // return;
  FlashCardInBoxType *flashCardInBoxP = 
    ctrlObjP->m_modelObjP->GetCurCard ();
  ctrlObjP->m_modelObjP->PrintCardInBoxInfo (flashCardInBoxP);
}


static void 
PresentCard (ControllerClass *ctrlObjP)
{
  GtkBuilder *builderP = ctrlObjP->m_builderP;

  /*
   * Re-present the new card content:
   */
  ctrlObjP->m_viewObjP->ClearPresentation ();
  ctrlObjP->m_viewObjP->Present ();// by presenting, will demote or promote
  ctrlObjP->StopSayCurrentWord (); // stop any pending sound


  /*
   * say the word if autoSound is enabled: 
   */
  const char *autoSound_str = 
    ctrlObjP->m_attributeObjP->GetValue ("autoSound_enum");
  if (strcmp (autoSound_str, AUTOSOUND_OFF) != 0)
    {
      /* say the word in Study Mode: */
      GtkNotebook *the_tabsP = 
        GTK_NOTEBOOK (gtk_builder_get_object (builderP, "the_tabs"));
      assert (the_tabsP);
      gint curTag_int = gtk_notebook_get_current_page (the_tabsP);
      
      /* say the word accordingly: */
      if ((curTag_int == ctrlObjP->m_frontPageNum_int) &&
          (strcmp (autoSound_str, AUTOSOUND_ON_FRONT) == 0))
        ctrlObjP->SayCurrentWord (); // say the word in the current card
      if ((curTag_int == ctrlObjP->m_backPageNum_int) &&
          (strcmp (autoSound_str, AUTOSOUND_ON_BACK) == 0))
        ctrlObjP->SayCurrentWord (); // say the word in the current card
    }
}

extern "C" G_MODULE_EXPORT void 
on_open_mItem_activate (GtkMenuItem *in_menuItem,
                        gpointer     in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  
  /*
   * Get the file selection dialog:
   */
  GtkWidget *dialog;
  dialog = GTK_WIDGET (gtk_builder_get_object (builderP, "fileOpen_dialog"));
  assert (dialog);

  /*
   * Set the defefault directory if the first time:
   */
  static bool firstTime_flg = true;
  if (unlikely (firstTime_flg))
    {
      firstTime_flg = false;
      // const char *homeDir_str = getenv ("HOME");
      // gchar *exampleDir_str = g_strdup_printf ("%s/MyDocs/gcardsys/", 
      // 					       homeDir_str);
      // chdir (homeDir_str);
      // gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), 
      // 					   exampleDir_str);
      // g_free (exampleDir_str);
      chdir (PKGDATADIR);	// came from gCardSys.pro file
      gtk_file_chooser_add_shortcut_folder (GTK_FILE_CHOOSER (dialog),
					    PKGDATADIR, NULL);
    }


  /* 
   * set file filters:
   */
  GtkFileFilter *cbx_filter = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (cbx_filter, "*.cbx");
  gtk_file_filter_set_name (cbx_filter, "CBX Only");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), cbx_filter);  
  gtk_object_destroy (GTK_OBJECT (cbx_filter));
  
  GtkFileFilter *all_filter = gtk_file_filter_new ();
  gtk_file_filter_add_pattern (all_filter, "*.*");
  gtk_file_filter_set_name (all_filter, "ALL");
  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), all_filter);  
  gtk_object_destroy (GTK_OBJECT (all_filter));


  /*
   * show the dialog with everything in it:
   */
  gtk_widget_show_all (dialog);
  gint dialogRet_int = gtk_dialog_run (GTK_DIALOG (dialog));


  /* 
   * Got response from user -- open the file:
   */
  if (dialogRet_int == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename;
      filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      fprintf (stderr, "full-pathed file name is %s\n", filename);
      modelObjP->Open (filename); // open the file
      g_free (filename);
    }


  /*
   * Clean dialog:
   */
  gtk_widget_hide (dialog);
  GSList *filter_listP = 
    gtk_file_chooser_list_filters (GTK_FILE_CHOOSER (dialog));
  cbx_filter = (GtkFileFilter*) (filter_listP->next->data);
  all_filter = (GtkFileFilter*) (filter_listP->data);
  gtk_file_chooser_remove_filter (GTK_FILE_CHOOSER (dialog), cbx_filter);
  gtk_file_chooser_remove_filter (GTK_FILE_CHOOSER (dialog), all_filter);
  g_slist_free (filter_listP);
  
  gtk_widget_show (dialog); // needed to get the filter right??
  gtk_widget_hide (dialog);


  /*
   * Bail out if the user did not choose a file:
   */
  if (dialogRet_int != GTK_RESPONSE_ACCEPT)
    return;                     // user canceled


  /*
   * Offer Test.
   */
  int neverVisited_int = 0;
  int imaturePcounts_int = 0;
  modelObjP->GetNumCards (NULL, NULL, NULL, NULL, NULL,
                          &neverVisited_int, &imaturePcounts_int, NULL, NULL);
  if ((neverVisited_int == 0) && (imaturePcounts_int == 0))
    {
      /* Asking to take a test or not: */
      GtkWidget *dialog = gtk_message_dialog_new 
        (NULL,
         GTK_DIALOG_MODAL,
         GTK_MESSAGE_QUESTION,
         GTK_BUTTONS_NONE,
         "Do you want to take a test now?");

      gtk_dialog_add_button (GTK_DIALOG (dialog),
                             GTK_STOCK_NO,
                             GTK_RESPONSE_NO);
      GtkWidget *yes_button = gtk_dialog_add_button (GTK_DIALOG (dialog),
                                                     GTK_STOCK_YES,
                                                     GTK_RESPONSE_YES);
      gtk_widget_grab_default (yes_button);
      gtk_widget_grab_focus (yes_button);
      gtk_widget_show (dialog);

      gint dialogRet_int = gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);  // clean up pop-up


      /* 
       * Got response from user:
       */
      if (dialogRet_int == GTK_RESPONSE_YES)
        {
          fprintf (stderr, "Answered Yes!\n");
          if (modelObjP->LoadTestSession (NULL) == false)
            {
              fprintf (stderr, "error loading Test session\n");

              /* Inform user test session failed loading: */
              GtkWidget *dialog = 
                gtk_message_dialog_new (NULL,
                                        GTK_DIALOG_MODAL,
                                        GTK_MESSAGE_INFO,
                                        GTK_BUTTONS_OK,
                                        "Test file is corrupted.");
              gtk_dialog_run (GTK_DIALOG (dialog));
              gtk_widget_destroy (dialog);  // clean up pop-up
            }

          /* set auto flip to be on front for testing mode: */
          GtkWidget *gtkWidgetP = 
            GTK_WIDGET (gtk_builder_get_object 
                        (builderP, "flipToFront_radiomItem"));
          g_signal_emit_by_name (G_OBJECT(gtkWidgetP), "activate"); 
        }
      else if (dialogRet_int == GTK_RESPONSE_NO)
        fprintf (stderr, "Answered No\n");
      else                      
        {
          // canceled, don't load anything
          modelObjP->Close (); 
          return;
        }
    }
  

  /*
   * Get the first study card:
   */
  if (modelObjP->GetNextCard () == NULL)
    return;                     // no card at all??
  PrintCurStudyCard (ctrlObjP);


  /*
   * Disables: self
   */
  gtk_widget_set_sensitive (GTK_WIDGET (in_menuItem), FALSE);


  /*
   * Enables: save_mItem, refresh_mItem, close_mItem,
   */
  GtkWidget *gtkWidgetP;
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "save_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                   "cancelIdle_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "close_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);


  /*
   * Additional Enables: next_btn, yepNext_btn, nopeNext_btn, flip_btn,
   * (yep_radio, nope_radio -- if handheld mode is not on), sound_btn,
   * sound_chkmItem, front_textV, back_textV, status_textV
   */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "yepNext_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "nopeNext_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "flip_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  /* yep_radio and nope_radio depends on the handheld_chkmItem */
  GtkCheckMenuItem *handheld_chkmItemP = GTK_CHECK_MENU_ITEM 
    (gtk_builder_get_object (builderP, "handheld_chkmItem"));
  assert (handheld_chkmItemP);
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "mark_vbox"));
  assert (gtkWidgetP);
  if (gtk_check_menu_item_get_active  (handheld_chkmItemP))
    gtk_widget_set_sensitive (gtkWidgetP, FALSE);
  else
    gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "sound_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "the_tabs"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                   "moveToTrash_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);


  /*
   * Need to reset reviewLastBoxOnly and review2ndBoxOnly after open new
   * file:
   */
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (builderP, 
                                        "reviewLastBoxOnly_chkmItem"));
  assert (gtkWidgetP);
  if (gtk_check_menu_item_get_active  (GTK_CHECK_MENU_ITEM (gtkWidgetP)))
    {
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), TRUE);
    }
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (builderP, 
                                        "review2ndBoxOnly_chkmItem"));
  assert (gtkWidgetP);
  if (gtk_check_menu_item_get_active  (GTK_CHECK_MENU_ITEM (gtkWidgetP)))
    {
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), TRUE);
    }


  /*
   * update the application state:
   */
  ctrlObjP->m_isOpened_flg = true;


  /*
   * Ready to present the first study card:
   */
  PresentCard (ctrlObjP);
}



extern "C" G_MODULE_EXPORT  void 
on_close_mItem_activate (GtkMenuItem *in_menuItem,
                         gpointer     in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  GtkWidget *gtkWidgetP;
  
  /*
   * Stop the playback
   */
  ctrlObjP->StopSayCurrentWord ();


  /*
   * Saving un-saved changes:
   */
  if (modelObjP->NeedSave ())
    {
      /* Asking to save or not: */
      GtkWidget *dialog = gtk_message_dialog_new 
        (NULL,
         GTK_DIALOG_MODAL,
         GTK_MESSAGE_QUESTION,
         GTK_BUTTONS_NONE,
         "Do you want to save the study/test session?");

      gtk_dialog_add_button (GTK_DIALOG (dialog),
                             GTK_STOCK_CANCEL,
                             GTK_RESPONSE_CANCEL);
      gtk_dialog_add_button (GTK_DIALOG (dialog),
                             GTK_STOCK_NO,
                             GTK_RESPONSE_NO);
      GtkWidget *yes_button = gtk_dialog_add_button (GTK_DIALOG (dialog),
                                                     GTK_STOCK_YES,
                                                     GTK_RESPONSE_YES);
      gtk_widget_grab_default (yes_button);
      gtk_widget_grab_focus (yes_button);
      gtk_widget_show (dialog);

      gint dialogRet_int = gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);  // clean up pop-up


      /* 
       * Got response from user:
       */

      int exportDkf_int = 
        atoi (ctrlObjP->m_attributeObjP->GetValue ("exportDkf"));
      if (dialogRet_int == GTK_RESPONSE_YES)
        {
          // don't save last card:
          if (modelObjP->Save (NULL, exportDkf_int, true) == false)
            {
              /* A visual feedback. */
              GtkWidget *dialog = 
                gtk_message_dialog_new (NULL,
                                        GTK_DIALOG_MODAL,
                                        GTK_MESSAGE_INFO,
                                        GTK_BUTTONS_OK,
                                        "Saving to file failed.");
              gtk_dialog_run (GTK_DIALOG (dialog));
              gtk_widget_destroy (dialog);  // clean up pop-up
              return;
            }
        }
      else if (dialogRet_int == GTK_RESPONSE_NO)
        fprintf (stderr, "No\n");
      else
        return;
    }


  /*
   * Disables: self, refresh_mItem, save_mItem, cancelIdle_mItem
   */
  gtk_widget_set_sensitive (GTK_WIDGET (in_menuItem), FALSE);

  // gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
  //                                                  "refresh_mItem"));
  // assert (gtkWidgetP);
  // gtk_widget_set_sensitive (gtkWidgetP, FALSE);
  // TC 2009-07-19 -- TBF

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "save_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, 
                                                   "cancelIdle_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);


  /*
   * Enables: open_mItem
   */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "open_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);


  /*
   * Additional Disables: next_btn, next_btn, flip_btn, yep_radio,
   * nope_radio, handHeldBack_btn, sound_btn, front_textV, back_textV,
   * status_textV (the_tabs), moveToTrash_mItem
   */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "prev_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "yepNext_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "nopeNext_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "flip_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object // mark_vbox includes
                           (builderP, "mark_vbox")); // yep_radio and nope_radio
  assert (gtkWidgetP);     
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object 
                           (builderP, "handHeldBack_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "sound_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "the_tabs"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);

  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object 
                           (builderP, "moveToTrash_mItem"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, FALSE);


  /*
   * Clear the presentation:
   */
  ctrlObjP->m_viewObjP->ClearPresentation ();


  /*
   * Close the model:
   */
  modelObjP->Close ();


  /*
   * update the application state:
   */
  ctrlObjP->m_isOpened_flg = false;
}


extern "C" G_MODULE_EXPORT  void 
on_save_mItem_activate (GtkMenuItem *in_menuItem,
                        gpointer     in_userData)
{
  fakeUseParam (in_menuItem, in_userData);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  
  int exportDkf_int = 
    atoi (ctrlObjP->m_attributeObjP->GetValue ("exportDkf"));

  /* explicit save will save the last card */
  if (modelObjP->Save (NULL, exportDkf_int, false) == false)
    {
      /* A visual feedback. */
      GtkWidget *dialog = 
        gtk_message_dialog_new (NULL,
                                GTK_DIALOG_MODAL,
                                GTK_MESSAGE_INFO,
                                GTK_BUTTONS_OK,
                                "Saving to file failed.");
      gtk_dialog_run (GTK_DIALOG (dialog));
      gtk_widget_destroy (dialog);  // clean up pop-up
    }
}


extern "C" G_MODULE_EXPORT  void 
on_cancelIdle_mItem_activate (GtkMenuItem *in_menuItem,
                            gpointer     in_userData)
{
  fakeUseParam (in_menuItem, in_userData);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ctrlObjP->m_modelObjP->SetPauseTimeOut (0); // immediate pause


  /*
   *  A visual feedback.
   */
  GtkWidget *dialog = 
    gtk_message_dialog_new (NULL,
                            GTK_DIALOG_MODAL,
                            GTK_MESSAGE_INFO,
                            GTK_BUTTONS_OK,
                            "Current idle timer canceled.");
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);  // clean up pop-up
}


extern "C" G_MODULE_EXPORT  void 
on_quit_mItem_activate (GtkMenuItem *in_menuItem,
                        gpointer     in_userData)
{
  fakeUseParam (in_menuItem, in_userData);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;

  GtkWidget *gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (builderP, "close_mItem"));
  assert (gtkWidgetP);
  on_close_mItem_activate ((GtkMenuItem*) gtkWidgetP, in_userData);

  if (ctrlObjP->m_isOpened_flg == false) // check to see if canceled
    gtk_main_quit ();
}



extern "C" G_MODULE_EXPORT  gboolean 
on_main_window_delete_event (GtkWidget *widget,
                             GdkEvent  *event,
                             gpointer   user_data)
{
  fakeUseParam (widget, event, user_data);
  fprintf (stderr, "called delete event.\n");


  ControllerClass *ctrlObjP = (ControllerClass*) user_data;
  GtkMenuItem *quit_mItemP = 
    GTK_MENU_ITEM (gtk_builder_get_object (ctrlObjP->m_builderP, "quit_mItem"));
  assert (quit_mItemP);
  on_quit_mItem_activate (quit_mItemP, user_data);

  gtk_main_quit ();             // TBF
  return FALSE;                 // not handled
}




extern "C" G_MODULE_EXPORT void
on_prev_btn_clicked (GtkButton *in_widget,
                     gpointer   in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  GtkBuilder *builderP = ctrlObjP->m_builderP;


  /*
   * Disable Prev if no more cards:
   */
  if (modelObjP->GetPrevCard () == NULL)
    {
      gtk_widget_set_sensitive (GTK_WIDGET (in_widget), FALSE);
      return;                     // no card at all??
    }
  
  /*
   * Enable Next if have more card:
   */
  GtkWidget *gtkWidgetP;
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);


  /*
   * Re-present the new card content:
   */
  PresentCard (ctrlObjP);
}


extern "C" G_MODULE_EXPORT void
on_handHeldBack_btn_clicked (GtkButton *in_widget,
                             gpointer   in_userData)
{
  on_prev_btn_clicked (in_widget, in_userData);
}


extern "C" G_MODULE_EXPORT void
on_flip_btn_clicked (GtkButton *in_widget,
                     gpointer   in_userData)
{
  fakeUseParam (in_widget);

  /*
   * Get the_tabs:
   */
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkNotebook *the_tabsP = 
    GTK_NOTEBOOK (gtk_builder_get_object (builderP, "the_tabs"));
  assert (the_tabsP);

  /*
   * switch between Front and Back tabs: 
   */
  gint curTag_int = gtk_notebook_get_current_page (the_tabsP);
  if (curTag_int == ctrlObjP->m_frontPageNum_int)
    gtk_notebook_set_current_page (the_tabsP, ctrlObjP->m_backPageNum_int);
  else if (curTag_int == ctrlObjP->m_backPageNum_int)
    gtk_notebook_set_current_page (the_tabsP, ctrlObjP->m_frontPageNum_int);




  /*
   * say the word if autoSound is enabled: 
   */
  ctrlObjP->StopSayCurrentWord (); // stop any pending sound
  const char *autoSound_str = 
    ctrlObjP->m_attributeObjP->GetValue ("autoSound_enum");
  if (strcmp (autoSound_str, AUTOSOUND_OFF) != 0)
    {
      GtkNotebook *the_tabsP = 
        GTK_NOTEBOOK (gtk_builder_get_object (builderP, "the_tabs"));
      assert (the_tabsP);
      gint curTag_int = gtk_notebook_get_current_page (the_tabsP);
      
      /* say the word accordingly: */
      if ((curTag_int == ctrlObjP->m_frontPageNum_int) &&
          (strcmp (autoSound_str, AUTOSOUND_ON_FRONT) == 0))
        ctrlObjP->SayCurrentWord (); // say the word in the current card
      if ((curTag_int == ctrlObjP->m_backPageNum_int) &&
          (strcmp (autoSound_str, AUTOSOUND_ON_BACK) == 0))
        ctrlObjP->SayCurrentWord (); // say the word in the current card
    }
}


extern "C" G_MODULE_EXPORT void
on_next_btn_clicked (GtkButton *in_widget,
                     gpointer   in_userData)
{
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget *gtkWidgetP;

  fprintf (stderr, "next clicked\n");

  /*
   * Disable Next if no more cards:
   */
  if (modelObjP->GetNextCard () == NULL)
    {
      gtk_widget_set_sensitive (GTK_WIDGET (in_widget), FALSE);

      return;                     // no card at all??
    }
  // PrintCurStudyCard (ctrlObjP);




  /*
   * Enable Prev if have more card:
   */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "prev_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);

  /* also enable the hand-held back_btn: */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object 
                           (builderP, "handHeldBack_btn"));
  assert (gtkWidgetP);
  gtk_widget_set_sensitive (gtkWidgetP, TRUE);


  /*
   * Re-present the new card content:
   */
  PresentCard (ctrlObjP);
}


extern "C" G_MODULE_EXPORT void
on_yepNext_btn_clicked (GtkButton *in_widget,
                        gpointer   in_userData)
{
  fakeUseParam (in_userData, in_widget);
  fprintf (stderr, "yepNext clicked\n");
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget *gtkWidgetP;

  /* do a "ype_radio" click: */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "yep_radio"));
  assert (gtkWidgetP);
  gtk_button_clicked (GTK_BUTTON (gtkWidgetP));


  /* do a "next_btn" click: */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  // on_next_btn_clicked (GTK_BUTTON (gtkWidgetP), in_userData);
  g_signal_emit_by_name (G_OBJECT(gtkWidgetP), 
                         "clicked"); // hildon use toolbutton
}




extern "C" G_MODULE_EXPORT void
on_nopeNext_btn_clicked (GtkButton *in_widget,
                         gpointer   in_userData)
{
  fakeUseParam (in_userData, in_widget);
  fprintf (stderr, "nopeNext clicked\n");
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  GtkWidget *gtkWidgetP;

  /* do a "nop_radio" click: */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "nope_radio"));
  assert (gtkWidgetP);
  gtk_button_clicked (GTK_BUTTON (gtkWidgetP));

  /* do a "next_btn" click: */
  gtkWidgetP = GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  // on_next_btn_clicked (GTK_BUTTON (gtkWidgetP), in_userData);
  g_signal_emit_by_name (G_OBJECT(gtkWidgetP), 
                         "clicked"); // hildon uses toolbutton
}


extern "C" G_MODULE_EXPORT void
on_nope_radio_toggled (GtkToggleButton *in_toggleButtonP,
                       gpointer         in_userData)
{
  fakeUseParam (in_toggleButtonP, in_userData);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;

  if (gtk_toggle_button_get_active (in_toggleButtonP))
    {
      fprintf (stderr, "Radio Demote\n");
      modelObjP->Demote ();

      // fprintf (stderr, "Demoted\n");
      // PrintCurStudyCard (ctrlObjP);
    }
  else
    {
      fprintf (stderr, "Radio Promote\n");
      modelObjP->Promote ();

      // fprintf (stderr, "Promoted\n");
      // PrintCurStudyCard (ctrlObjP);
    }
}


extern "C" G_MODULE_EXPORT void
on_sound_btn_clicked (GtkButton *in_widget,
                      gpointer   in_userData)
{
  fakeUseParam (in_widget);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;

  /* Ask a timer to repeatedly say the word: */
  static bool soundStarting_flg = false; // prevent user clicking too fast
  if (soundStarting_flg == false &&
      ctrlObjP->m_repeatTimerId_uint == 0)
    {
      soundStarting_flg = true;
      ctrlObjP->StopSayCurrentWord ();
      ctrlObjP->SayCurrentWord (); 
      ctrlObjP->m_repeatTimerId_uint = 
        // g_timeout_add (50, (GSourceFunc) 
        g_timeout_add (150, (GSourceFunc) 
                       ctrlObjP->RepeatSayCurrentWord, 
                       ctrlObjP);
      soundStarting_flg = false;
    }
  else
    {
      /* toggle the continuous playback to stop: */
      // g_source_remove (ctrlObjP->m_repeatTimerId_uint);
      // ctrlObjP->m_repeatTimerId_uint = 0;

      // if (g_main_loop_is_running (ctrlObjP->m_gstMainLoopP))
      //   g_main_loop_quit (ctrlObjP->m_gstMainLoopP); 

      ctrlObjP->StopSayCurrentWord ();
    }
}



/**
 *
 * Handle hardware key presses. Currently will switch into and out of
 * fullscreen mode if the "fullscreen" button is pressed (F6 in the
 * SDK).
 *
 * As the keypresses come from outside GTK+ (and even GDK), this
 * needs to be an event handler.
 */
extern "C" G_MODULE_EXPORT gboolean 
on_main_window_key_press_event (GtkWidget   *in_widget,
                                GdkEventKey *in_event,
                                gpointer     in_userData)
{
  fakeUseParam (in_userData, in_widget);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  fakeUseParam (builderP);

  fprintf (stderr, "key pressed %d\n", in_event->keyval);

  switch (in_event->keyval) 
    {
#ifdef IS_HILDON
      /* see ./hildon-1/hildon/hildon-defines.h */
    case HILDON_HARDKEY_FULLSCREEN:
      {
	fprintf(stderr, " Fullscreen hw-button pressed (or F6 in SDK)\n");

	GtkWindow *main_windowP = 
	  GTK_WINDOW (gtk_builder_get_object (builderP, "main_window"));

	/* Toggle fullscreen mode. */
	if (ctrlObjP->m_isFullScreen_flg) 
	  {
	    gtk_window_unfullscreen (main_windowP);
	    ctrlObjP->m_isFullScreen_flg = false;
	  } 
	else
	  {
	    gtk_window_fullscreen (main_windowP);
	    ctrlObjP->m_isFullScreen_flg = true;
	  }
	return TRUE;
      }
#endif
      /* see /usr/include/gtk-2.0/gdk/gdkkeysyms.h */
    case GDK_Return:            // also HILDON_HARDKEY_SELECT
      {
        FlashCardInBoxType *flashCardInBoxP = 
          ctrlObjP->m_modelObjP->GetCurCard ();
        if (flashCardInBoxP)
          gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_PRIMARY),
                                  flashCardInBoxP->flashCardP->front_str,
                                  -1);
      }
      break;

    default:
      ;
  }

  /* We didn't process the event. */
  return FALSE;
}


extern "C" G_MODULE_EXPORT  void 
on_moveToTrash_mItem_activate (GtkMenuItem *in_menuItem,
                            gpointer     in_userData)
{
  fakeUseParam (in_menuItem, in_userData);
  ControllerClass *ctrlObjP = (ControllerClass*) in_userData;
  ModelClass *modelObjP = ctrlObjP->m_modelObjP;
  GtkBuilder *builderP = ctrlObjP->m_builderP;
  
  modelObjP->Trash ();

  /* do a "next_btn" click: */
  GtkWidget *gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (builderP, "next_btn"));
  assert (gtkWidgetP);
  // on_next_btn_clicked (GTK_BUTTON (gtkWidgetP), in_userData);
  g_signal_emit_by_name (G_OBJECT(gtkWidgetP), 
                         "clicked"); // hildon uses toolbutton
}



/*************************
 * ControllerClass:
 ************************/


ControllerClass::
ControllerClass (ViewClass      *in_viewObjP,
                 AttributeClass *in_attributeObjP)
{
  m_viewObjP = in_viewObjP;
  m_modelObjP = in_viewObjP->m_modelObjP;
  m_attributeObjP = in_attributeObjP;
  assert (m_attributeObjP);     // TBF

  m_repeatTimerId_uint = 0;
  m_gstMainLoopP = g_main_loop_new (NULL, FALSE);
  m_esd_fd = 0;                 // for Maemo N8x0 only

  m_isFullScreen_flg = false;
  m_isOpened_flg = false;
  m_frontPageNum_int = -1;
  m_backPageNum_int = -1;
  m_soundError_flg = false;
}


ControllerClass::
~ControllerClass ()
{
  g_main_loop_unref (m_gstMainLoopP);

}



void ControllerClass::
Init (GtkBuilder *in_builderP,
      const char *in_title_str)
{
  m_builderP = in_builderP;
  fprintf (stderr, "Calling Init\n");

  /*
   * Set title:
   */
  GtkWidget *main_windowP = 
    GTK_WIDGET (gtk_builder_get_object (m_builderP, "main_window"));
  assert (main_windowP);
  gtk_window_set_title (GTK_WINDOW (main_windowP), in_title_str);


#ifdef IS_HILDON
  /*
   * Re-root the main window for Hilton:
   */
  HildonProgram* program;
  HildonWindow* window;

  /* Create the Hildon program. */
  program = HILDON_PROGRAM (hildon_program_get_instance ());

  /* Set the application title using an accessor function. */
  g_set_application_name (in_title_str);

  /* Create a window that will handle our layout and menu. */
  window = HILDON_WINDOW (hildon_window_new ());

  /* Bind the HildonWindow to HildonProgram. */
  hildon_program_add_window (program, HILDON_WINDOW (window));

  /* Connect the termination signals. */
  g_signal_connect (G_OBJECT(window), "delete-event",
                    G_CALLBACK (on_main_window_delete_event), this);
  g_signal_connect (G_OBJECT(window), "destroy-event",
                    G_CALLBACK (on_main_window_destroy_event), this);

  /* reroute the content_vbox */
  GtkWidget *content_vboxP = 
    GTK_WIDGET (gtk_builder_get_object (m_builderP, "content_vbox"));
  assert (content_vboxP);
  g_object_set_data (G_OBJECT (content_vboxP), "main_window", window);

  g_object_ref (content_vboxP);
  gtk_container_remove (GTK_CONTAINER (main_windowP), content_vboxP);
  gtk_container_add (GTK_CONTAINER (window), content_vboxP);


  main_windowP = GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                                     "main_window"));

//   g_object_unref (main_windowP); 
// TC 2009-07-19 -- cause GLIB CRITICAL 


  /*
   * Re-construct the menu for hildon:
   */

  /* 
   * Create a autoFlip sub-menu: 
   */
  GtkWidget *autoFlip_subMenuP = gtk_menu_new ();
  GSList *flipRdoGroupP = NULL;
  GtkWidget* flipToFront_radiomItemP = 
    gtk_radio_menu_item_new_with_label (flipRdoGroupP, "to front (study)");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoFlip_subMenuP), 
                         flipToFront_radiomItemP);
  flipRdoGroupP = gtk_radio_menu_item_get_group 
    (GTK_RADIO_MENU_ITEM (flipToFront_radiomItemP));
  GtkWidget* flipToBack_radiomItemP = 
    gtk_radio_menu_item_new_with_label (flipRdoGroupP, "to back (review)");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoFlip_subMenuP), 
                         flipToBack_radiomItemP);
  flipRdoGroupP = gtk_radio_menu_item_get_group 
    (GTK_RADIO_MENU_ITEM (flipToBack_radiomItemP));
  GtkWidget* flipOff_radiomItemP = 
    gtk_radio_menu_item_new_with_label (flipRdoGroupP, "off");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoFlip_subMenuP), 
                         flipOff_radiomItemP);

  /* tie the autoFlip sub-menu to autoFlip menu item: */
  GtkWidget* autoFlip_chkmItemP = 
    gtk_menu_item_new_with_label ("AutoFlip");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (autoFlip_chkmItemP), 
                             autoFlip_subMenuP);

  /* 
   * Create a autoSound sub-menu: 
   */
  GtkWidget *autoSound_subMenuP = gtk_menu_new ();
  GSList *soundRdoGroupP = NULL;
  GtkWidget* soundOnFront_radiomItemP = 
    gtk_radio_menu_item_new_with_label (soundRdoGroupP, "on front (review)");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoSound_subMenuP), 
                         soundOnFront_radiomItemP);
  soundRdoGroupP = gtk_radio_menu_item_get_group 
    (GTK_RADIO_MENU_ITEM (soundOnFront_radiomItemP));
  GtkWidget* soundOnBack_radiomItemP = 
    gtk_radio_menu_item_new_with_label (soundRdoGroupP, "on back (study)");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoSound_subMenuP), 
                         soundOnBack_radiomItemP);
  soundRdoGroupP = gtk_radio_menu_item_get_group 
    (GTK_RADIO_MENU_ITEM (soundOnBack_radiomItemP));
  GtkWidget* soundOff_radiomItemP = 
    gtk_radio_menu_item_new_with_label (soundRdoGroupP, "off");
  gtk_menu_shell_append (GTK_MENU_SHELL (autoSound_subMenuP), 
                         soundOff_radiomItemP);

  /* tie the autoSound sub-menu to autoSound menu item: */
  GtkWidget* autoSound_chkmItemP = 
    gtk_menu_item_new_with_label ("AutoSound");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (autoSound_chkmItemP), 
                             autoSound_subMenuP);

  /*
   * Create an option sub-menu: 
   */
  GtkWidget* option_subMenuP = gtk_menu_new ();
  GtkWidget* handheld_chkmItemP = 
    gtk_check_menu_item_new_with_label ("Handheld");
  gtk_menu_shell_append (GTK_MENU_SHELL (option_subMenuP), 
                         handheld_chkmItemP);
  GtkWidget* reviewLastBoxOnly_chkmItemP = 
    gtk_check_menu_item_new_with_label ("ReviewLastBoxOnly");
  gtk_menu_shell_append (GTK_MENU_SHELL (option_subMenuP), 
                         reviewLastBoxOnly_chkmItemP);
  GtkWidget* review2ndBoxOnly_chkmItemP = 
    gtk_check_menu_item_new_with_label ("Review2ndBoxOnly");
  gtk_menu_shell_append (GTK_MENU_SHELL (option_subMenuP), 
                         review2ndBoxOnly_chkmItemP);
  gtk_container_add (GTK_CONTAINER (option_subMenuP), autoFlip_chkmItemP);
  gtk_container_add (GTK_CONTAINER (option_subMenuP), autoSound_chkmItemP);

  /* tie the option sub-menu to option menu item: */
  GtkWidget* miOptions = gtk_menu_item_new_with_label ("Option");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (miOptions), option_subMenuP);


  /*
   * Create a speicalOp sub-menu: 
   */
  GtkWidget* specialOp_subMenuP = gtk_menu_new ();
  GtkWidget* moveToTrash_mItemP = 
    gtk_menu_item_new_with_label ("Move to Trash CardBox");
  gtk_menu_shell_append (GTK_MENU_SHELL (specialOp_subMenuP), 
                         moveToTrash_mItemP);

  /* tie speicalOp sub-menu to specialOp menu item: */
  GtkWidget* miSpecialOps = gtk_menu_item_new_with_label ("SpecialOp");
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (miSpecialOps), specialOp_subMenuP);


  /*
   * Create a top-level new menu:
   */
  GtkMenu* menu = GTK_MENU (gtk_menu_new ());
  GtkWidget *miOpen = gtk_menu_item_new_with_label("Open");
  GtkWidget *miSave = gtk_menu_item_new_with_label("Save");
  GtkWidget *miCancelIdle = gtk_menu_item_new_with_label("Cancel Idle");
  GtkWidget *miClose = gtk_menu_item_new_with_label("Close");
  GtkWidget *miSep1 = gtk_separator_menu_item_new();
  GtkWidget *miSep2 = gtk_separator_menu_item_new();
  GtkWidget *miQuit = gtk_menu_item_new_with_label("Quit");


  /* Add the items to the top-level menu: */
  gtk_container_add (GTK_CONTAINER (menu), miOpen);
  gtk_container_add (GTK_CONTAINER (menu), miClose);
  gtk_container_add (GTK_CONTAINER (menu), miSep1);
  gtk_container_add (GTK_CONTAINER (menu), miSave);
  gtk_container_add (GTK_CONTAINER (menu), miCancelIdle);
  gtk_container_add (GTK_CONTAINER (menu), miOptions);
  gtk_container_add (GTK_CONTAINER (menu), miSpecialOps);
  gtk_container_add (GTK_CONTAINER (menu), miSep2);
  gtk_container_add (GTK_CONTAINER (menu), miQuit);

  /* Set the top-level menu for Hildon program . */
  hildon_program_set_common_menu (program, menu);


  /*
   * Connect the signals from the individual menu items. 
   */
  g_signal_connect (G_OBJECT (miOpen), "activate",
    G_CALLBACK (on_open_mItem_activate), this);
  g_signal_connect (G_OBJECT (miSave), "activate",
                    G_CALLBACK (on_save_mItem_activate), this);
  g_signal_connect (G_OBJECT (miCancelIdle), "activate",
                    G_CALLBACK (on_cancelIdle_mItem_activate), this);
  g_signal_connect (G_OBJECT (miClose), "activate",
                    G_CALLBACK (on_close_mItem_activate), this);
  g_signal_connect (G_OBJECT (miQuit), "activate",
                    G_CALLBACK (on_quit_mItem_activate), this);
  g_signal_connect (G_OBJECT (handheld_chkmItemP), "activate",
                    G_CALLBACK (on_handheld_chkmItem_toggled), this);
  g_signal_connect (G_OBJECT (flipToFront_radiomItemP), "activate",
                    G_CALLBACK (on_flipToFront_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (flipToBack_radiomItemP), "activate",
                    G_CALLBACK (on_flipToBack_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (flipOff_radiomItemP), "activate",
                    G_CALLBACK (on_flipOff_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (soundOnFront_radiomItemP), "activate",
                    G_CALLBACK (on_soundOnFront_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (soundOnBack_radiomItemP), "activate",
                    G_CALLBACK (on_soundOnBack_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (soundOff_radiomItemP), "activate",
                    G_CALLBACK (on_soundOff_radiomItem_toggled), this);
  g_signal_connect (G_OBJECT (reviewLastBoxOnly_chkmItemP), "activate",
                    G_CALLBACK (on_reviewLastBoxOnly_chkmItem_toggled), this);
  g_signal_connect (G_OBJECT (review2ndBoxOnly_chkmItemP), "activate",
                    G_CALLBACK (on_review2ndBoxOnly_chkmItem_toggled), this);
  g_signal_connect (G_OBJECT (moveToTrash_mItemP), "activate",
                    G_CALLBACK (on_moveToTrash_mItem_activate), this);

  /*
   * Re-connect signal for nope_radio (bug in GtkBuilder 2.10?):
   */
  GtkWidget *nope_radioP = 
    GTK_WIDGET (gtk_builder_get_object (m_builderP, "nope_radio"));
  assert (nope_radioP);
  g_signal_connect (G_OBJECT (nope_radioP), "toggled",
                    G_CALLBACK (on_nope_radio_toggled), this);

  /* 
   * replacement to widgets: 
   */
  g_object_set_data (G_OBJECT (content_vboxP), "open_mItem", miOpen);
  g_object_set_data (G_OBJECT (content_vboxP), "save_mItem", miSave);
  g_object_set_data (G_OBJECT (content_vboxP), "cancelIdle_mItem", 
                     miCancelIdle);
  g_object_set_data (G_OBJECT (content_vboxP), "close_mItem", miClose);
  g_object_set_data (G_OBJECT (content_vboxP), "quit_mItem", miQuit);
  g_object_set_data (G_OBJECT (content_vboxP), "handheld_chkmItem",
                     handheld_chkmItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "flipToFront_radiomItem",
                     flipToFront_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "flipToBack_radiomItem",
                     flipToBack_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "flipOff_radiomItem",
                     flipOff_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "soundOnFront_radiomItem",
                     soundOnFront_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "soundOnBack_radiomItem",
                     soundOnBack_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "soundOff_radiomItem",
                     soundOff_radiomItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "reviewLastBoxOnly_chkmItem",
                     reviewLastBoxOnly_chkmItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "review2ndBoxOnly_chkmItem",
                     review2ndBoxOnly_chkmItemP);
  g_object_set_data (G_OBJECT (content_vboxP), "moveToTrash_mItem",
                     moveToTrash_mItemP);

  /*
   * Show the hildon menu:
   */
  gtk_widget_show_all (GTK_WIDGET (menu));
#endif


  /*
   * add key binding: (Need to re-get main_windowP in case HILDON)
   */
  main_windowP = GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                                     "main_window"));
  assert (main_windowP);
  fprintf (stderr, "routing keyevy to window %p\n", main_windowP);
  g_signal_connect (G_OBJECT (main_windowP), "key_press_event",
                    G_CALLBACK (on_main_window_key_press_event), this);


  /*
   * Setup one-time initialization for gStreamer
   */
  gst_init (NULL, NULL);


  /*
   * Setup one-time initialization for View:
   */
  m_viewObjP->InitGUI (in_builderP);


  /*
   * Default states for handheld_chkmItem
   */
  GtkWidget *gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (in_builderP, "handheld_chkmItem"));
  assert (gtkWidgetP);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), TRUE);


  /*
   * Default state for autoFlip
   */
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (in_builderP, "flipToBack_radiomItem"));
  assert (gtkWidgetP);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), TRUE);
  m_attributeObjP->SetValue ("autoFlip_enum", AUTOFLIP_TO_BACK);


  /*
   * Default state for autoSound
   */
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (in_builderP, "soundOff_radiomItem"));
  assert (gtkWidgetP);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), TRUE);
  m_attributeObjP->SetValue ("autoSound_enum", AUTOSOUND_OFF);


  /*
   * Default state for reviewLastBoxOnly and review2ndBoxOnly
   */
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (in_builderP, 
                                        "reviewLastBoxOnly_chkmItem"));
  assert (gtkWidgetP);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);
  gtkWidgetP = 
    GTK_WIDGET (gtk_builder_get_object (in_builderP, 
                                        "review2ndBoxOnly_chkmItem"));
  assert (gtkWidgetP);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (gtkWidgetP), FALSE);


  /*
   * The tabs: (get page numbers of back and front sides of the card)
   */
  GtkNotebook *the_tabsP = 
    GTK_NOTEBOOK (gtk_builder_get_object (m_builderP, "the_tabs"));
  assert (the_tabsP);
  GtkWidget *backCard_scrldWindowP =
    GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                        "backCard_scrldWindow"));
  assert (backCard_scrldWindowP);
  m_backPageNum_int = gtk_notebook_page_num  (the_tabsP, 
                                              backCard_scrldWindowP);
  GtkWidget *frontCard_textVP =
    GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                        "frontCard_textV"));
  assert (frontCard_textVP);
  m_frontPageNum_int = gtk_notebook_page_num  (the_tabsP, 
                                               frontCard_textVP);

  /*
   * Show the main window:
   */
  gtk_widget_show (main_windowP);
}


gboolean ControllerClass::
gStreamerBus_call (GstBus     *bus,
                   GstMessage *msg,
                   gpointer    data)
{
  fakeUseParam (bus);
  ControllerClass *ctrlObjP = (ControllerClass*) data;

  switch (GST_MESSAGE_TYPE (msg)) 
    {
    case GST_MESSAGE_EOS:
      g_main_loop_quit (ctrlObjP->m_gstMainLoopP);
      break;
      
    case GST_MESSAGE_ERROR:
      {
        gchar  *debug;
        GError *error;
        
        gst_message_parse_error (msg, &error, &debug);
        g_free (debug);

        g_printerr ("Error: %s\n", error->message);
        g_error_free (error);
        ctrlObjP->m_soundError_flg = true;
        
        if (ctrlObjP->m_gstMainLoopP)
          g_main_loop_quit (ctrlObjP->m_gstMainLoopP);
        break;
      }
    default:
      break;
    }

  return TRUE;
}


static void
on_pad_added (GstElement *element,
              GstPad     *pad,
              gpointer    data)
{
  fakeUseParam (element);
//   GstPad *sinkpad;
//   GstElement *decoder = (GstElement *) data;

//   /* We can now link this pad with the vorbis-decoder sink pad */
//   sinkpad = gst_element_get_static_pad (decoder, "sink");
//   gst_pad_link (pad, sinkpad);

//   /* clean ups: */
//   gst_object_unref (sinkpad);

  
  GstPad *sinkpad;
  GstElement *converter = (GstElement *) data;


  /* We can now link this pad with the converter sink pad */
  sinkpad = gst_element_get_static_pad (converter, "sink");
  gst_pad_link (pad, sinkpad);

  /* clean ups: */
  gst_object_unref (sinkpad);
}


#ifdef IS_HILDON
/* TC 2010-05-09 (Sun) -- copied from play-sound.c of osso-af-utils: */
/* Copypaste from esdfile.c and xmms mixer.c with (very) minor alterations. */
/* Plays a file with esd and with scaled channels. */
static bool
esd_play_file_with_pan (int         esd,
                        const char *name_prefix,
                        const char *filename, 
                        int         left_scale, 
                        int         right_scale)
{
  /* input from libaudiofile... */
  AFfilehandle in_file;
  int in_format, in_width, in_channels, frame_count;
  double in_rate;
  int bytes_per_frame;
  
  /* output to esound... */
  int out_sock, out_bits, out_channels, out_rate;
  int out_mode = ESD_STREAM, out_func = ESD_PLAY;
  esd_format_t out_format;
  char name[ ESD_NAME_MAX ] = "";
  
  esd_player_info_t *info;
  esd_info_t *all_info = NULL;
  
  /* open the audio file */
  in_file = afOpenFile( filename, "rb", NULL );
  if ( !in_file )
    return false;
  
  /* get audio file parameters */
  frame_count = afGetFrameCount( in_file, AF_DEFAULT_TRACK );
  in_channels = afGetChannels( in_file, AF_DEFAULT_TRACK );
  in_rate = afGetRate( in_file, AF_DEFAULT_TRACK );
  afGetSampleFormat( in_file, AF_DEFAULT_TRACK, &in_format, &in_width );
  
  if(getenv("ESDBG")) 
    {
      printf ("frames: %i channels: %i rate: %f format: %i width: %i\n",
              frame_count, in_channels, in_rate, in_format, in_width);
    }
  
  /* convert audiofile parameters to EsounD parameters */
  if ( in_width == 8 )
    out_bits = ESD_BITS8;
  else if ( in_width == 16 )
    out_bits = ESD_BITS16;
  else
    return false;
  
  bytes_per_frame = ( in_width  * in_channels ) / 8;
  
  if ( in_channels == 1 )
    out_channels = ESD_MONO;
  else if ( in_channels == 2 )
    out_channels = ESD_STEREO;
  else
    return false;
  
  out_format = out_bits | out_channels | out_mode | out_func;
  out_rate = (int) in_rate;
  
  /* construct name */
  if ( name_prefix ) 
    {
      strncpy( name, name_prefix, ESD_NAME_MAX - 2 );
      strcat( name, ":" );
    }
  strncpy( name + strlen( name ), filename, ESD_NAME_MAX - strlen( name ) );
  
  /* connect to server and play stream */
  /* fprintf (stderr, "filename is '%s'\n", filename); */
  out_sock = esd_play_stream( out_format, out_rate, NULL, filename );
  
  if ( out_sock <= 0 ) {
    /* Close the AF filehandle too */
    afCloseFile ( in_file );
    return false;
  }
  
  all_info = esd_get_all_info(esd);
  for (info = all_info->player_list; info != NULL; info = info->next) 
    {
      if (!strcmp(filename, info->name)) 
        {
          break;
        }
    }
  if (info != NULL) 
    {
      esd_set_stream_pan(esd, info->source_id,
                         left_scale, right_scale);
    }
  esd_free_all_info(all_info);
  
  /* play */
  esd_send_file (out_sock, in_file, bytes_per_frame);
  
  /* close up and go home */
  close (out_sock);
  if (afCloseFile (in_file))
    return false;
  
  return true;
}

static int 
is_esd_server_quiesce (int esd_fd)
{
  esd_server_info_t *esd_server_info = NULL ;
  int monitor_fd = -1, available_bytes = 0;
  if (NULL != (esd_server_info = esd_get_server_info (esd_fd)))
    {
      if ((monitor_fd = esd_monitor_stream (esd_server_info->format, 
                                            esd_server_info->rate, 
                                            NULL, NULL)) > 0)
        {
          ioctl (monitor_fd, FIONREAD, &available_bytes) ;

          struct timeval tv;
          tv.tv_sec = 0;
          tv.tv_usec = 200000;
          select (0, NULL, NULL, NULL, &tv); /* avoid tight loop */
          

          ioctl (monitor_fd, FIONREAD, &available_bytes) ;
          close (monitor_fd) ;
          esd_free_server_info (esd_server_info);
          return (available_bytes == 0);
        }
    }
  esd_free_server_info (esd_server_info);
  return 1;
}


/* TC 2010-05-09 (Sun) -- copied from play-sound.c of osso-af-utils with
   some minor reformating: */
static void
wait_for_esd_server_to_quiesce (int esd_fd)
{
  char *buf = NULL ;
  esd_server_info_t *esd_server_info = NULL ;
  int monitor_fd = -1, read_return = -1, available_bytes = 0, max_avail = 0;
  int first_time = 1;
  
  if (NULL != (esd_server_info = esd_get_server_info (esd_fd)))
    {
      if ((monitor_fd = esd_monitor_stream (esd_server_info->format, 
                                            esd_server_info->rate, 
                                            NULL, NULL)) > 0)
        {
          while (TRUE)
            {
              struct timeval tv;
              
              ioctl (monitor_fd, FIONREAD, &available_bytes) ;
              if (available_bytes > max_avail)
                {
                  buf = (char*) realloc (buf, available_bytes) ;
                  max_avail = available_bytes ;
                }
              
              if (available_bytes)
		{
                  read_return = read(monitor_fd, buf, available_bytes);
                  if (read_return < 0)
                    break ;
		}
              else if ( !first_time )
                {
                  break ;
                }
              
              tv.tv_sec = 0;
              tv.tv_usec = 200000;
              select (0, NULL, NULL, NULL, &tv); /* avoid tight loop */
              first_time = 0;
            }

          if (buf)
            free (buf) ;
          close (monitor_fd) ;
        }
      esd_free_server_info (esd_server_info) ;
    }
}
#endif


bool ControllerClass::
PlaySound (const char *in_soundFile_str)
{
#ifdef IS_HILDON
  /* Use the ESD method for much more responsive sound playback: */
  // int pan;
  // pan = 2;
  // pan *= SCALE_MULTIPLIER;
  int pan = 0xEC;               // bigger then E4 and smaller than F0
  m_esd_fd = esd_open_sound (NULL);
  return esd_play_file_with_pan (m_esd_fd, NULL, in_soundFile_str, pan, pan);

#else
  if (g_main_loop_is_running (m_gstMainLoopP))
    {
      fprintf (stderr, "gmain loop is already running.\n");
      return false;               // only one instance of mainloop allowed
    }
  
  /* we add pipe line and message handler */
  static GstElement *pipeline = gst_pipeline_new ("audio-player");

  /* we set the input filename to the source element */
  static GstElement *source = gst_element_factory_make ("filesrc", 
                                                        "file-source");

  /* setup the sound pipe line once only: */
  static bool firstTime_flg = true;
  if (unlikely (firstTime_flg))
    {
      firstTime_flg = false;

      GstBus *bus = gst_pipeline_get_bus (GST_PIPELINE (pipeline));
      gst_bus_add_watch (bus, gStreamerBus_call, this);
      gst_object_unref (bus);       // cleanup bus 

      /*
       * Use Wave: wavparse ! audioconvert ! alsasink
       */ 
      fprintf (stderr, "setting up gstream pipe line\n");
      static GstElement *decoder  = 
        gst_element_factory_make ("wavparse", "wav-parser");
      static GstElement *conv     = 
        gst_element_factory_make ("audioconvert", "converter");
// #ifdef IS_HILDON
//       static GstElement *sink     = 
//         gst_element_factory_make ("dsppcmsink", "audio-output");
// #else
      static GstElement *sink     = 
        gst_element_factory_make ("alsasink",  "audio-output");
// #endif
      gst_bin_add_many (GST_BIN (pipeline),
                        source, decoder, conv, sink, NULL);
      if (gst_element_link (source, decoder) == FALSE)
        {
          fprintf (stderr, "1. link pc failed\n");
          return false;
        }
      if (gst_element_link (conv, sink) == FALSE)
        {
          fprintf (stderr, "2. link pc failed\n");
          return false;
        }
      g_signal_connect (decoder, "pad-added", 
                        G_CALLBACK (on_pad_added), conv); 
    }

  /* set/change the sound source :*/
  g_object_set (G_OBJECT (source), "location", in_soundFile_str, NULL);

  /* test to see if file is there: */
  static struct stat _stat; // static will zero out
  if (stat (in_soundFile_str, &_stat) != 0)
    {
      fprintf (stderr, "sound file %s not found \n", in_soundFile_str);
      return false;
    }

  /* Set the pipeline to "playing" state: */
  GstStateChangeReturn changeState_enum =
    gst_element_set_state (pipeline, GST_STATE_PLAYING);
  if (changeState_enum == GST_STATE_CHANGE_FAILURE)
    {
      fprintf (stderr, "sound file %s could not set to play \n", 
               in_soundFile_str);
      return false;
    }

  /* Iterate */
  assert (g_main_loop_is_running (m_gstMainLoopP) == FALSE);
  g_main_loop_run (m_gstMainLoopP); // block until g_main_loop_quit()

  /* Out of the main loop, clean up nicely */
  gst_element_set_state (pipeline, GST_STATE_NULL);

  /* check error: */
  if (m_soundError_flg == true)
    {
      fprintf (stderr, "reset sound error\n");
      m_soundError_flg = false;
      return false;
    }

  return true;
#endif
}


gboolean ControllerClass::
RepeatSayCurrentWord (ControllerClass *in_ctrlObjP)
{
  // fprintf (stderr, "calling RepeatSayCurrentWord\n");
#ifdef IS_HILDON
  // wait_for_esd_server_to_quiesce (in_ctrlObjP->m_esd_fd);
  if (in_ctrlObjP->m_esd_fd &&
      is_esd_server_quiesce (in_ctrlObjP->m_esd_fd))
    {
      esd_close (in_ctrlObjP->m_esd_fd);
      in_ctrlObjP->m_esd_fd = 0;
      in_ctrlObjP->SayCurrentWord (); 
    }
  while (g_main_context_iteration(NULL, FALSE)); /* more responsive */
#else
  /* check to see ready to say the word again: */
  if (g_main_loop_is_running (in_ctrlObjP->m_gstMainLoopP) == FALSE)
    in_ctrlObjP->SayCurrentWord (); 
  
#endif
  return TRUE;      // continue timer, call StopSayCurrentWord () to stop
} 

static void
ExtractWord (const char *front_str,
             char       *word_str)
{
  /* word may stop at '*':*/
  const char *stop_chrP;
  stop_chrP = strstr (front_str, "*");
  if (stop_chrP)
    sscanf (front_str, "%[^*]", word_str);
  else
    {
      /* word may stop at the category enclosed in '()' */
      stop_chrP = strstr (front_str, "(");
      if (stop_chrP)
        sscanf (front_str, "%[^(]", word_str);
      else
        /* word may stop '\0' */
        sscanf (front_str, "%s", word_str);
    }
  
  /* Remove trailing spaces, in case: */
  for (int idx = strlen (word_str) - 1; idx >= 0; idx --)
    {
      if (word_str [idx] == ' ')
        {
          word_str [idx] = '\0';
          break;
        }
    }
}


void ControllerClass::
SayCurrentWord ()
{
  /*
   * Truncate the rest of the word starting from a non-letter position:
   * (This is needed because sometimes there is a '*' mark or part of
   * speech info) 
   *
   * TC 2010-07-17 (Sat) -- however, we now have spaces in words like "a
   * capella", which is represented as "a_capella".  So we need to allow
   * '_' as the part of the word sound file name.
   * TC 2010-07-31 (Sat) -- Also, we have dash '-' as in avant-garde
   * TC 2010-08-01 (Sun) -- convert ' ' to '_'
   */
  FlashCardInBoxType *cardInBox1P =  m_modelObjP->GetCurCard ();
  char word_str[80];        // TBF no word is 80 or greater!
  ExtractWord (cardInBox1P->flashCardP->front_str, word_str);

  /* replace ' ' with '_' for sound file name: */
  for (int idx = 0; idx < (int) strlen (word_str); idx ++)
    {
      if (word_str [idx] == ' ')
        word_str [idx] = '_';
    }

  /* 
   * Construct the entire path to sound file: (assume wav)
   */
  const char* fullSoundPath_str = cardInBox1P->deckObjP->GetFullSoundPath ();
  gchar *soundFile_str;

  /* convert to lower case: */
  for (int idx = 0; idx < strlen (word_str); idx++)
    word_str [idx] = tolower (word_str [idx]);

  if (fullSoundPath_str)
    soundFile_str = g_strdup_printf ("%s/%c/%s.wav", 
                                     fullSoundPath_str, 
                                     word_str[0],
                                     word_str);
  else
    soundFile_str = g_strdup_printf ("%s.wav", word_str);



  /*
   * Play the sound :
   */
  /* change current directory: */
  char *curDir_str = get_current_dir_name (); 
  assert (curDir_str);
  int retVal = chdir (m_modelObjP->GetCardBoxCabinetDir ());
  assert (retVal == 0);

  if (PlaySound (soundFile_str) == false) // should be blocking ?
    {
      fprintf (stderr, "PlaySound failed\n");
      StopSayCurrentWord ();
    }

  /* restore the current directory: */
  retVal = chdir (curDir_str);
  free (curDir_str);
  assert (retVal == 0);


  /*
   * Clean ups:
   */
  g_free (soundFile_str);
}


void ControllerClass::
StopSayCurrentWord ()
{
  /* stop existing repeat timer: */
  if (m_repeatTimerId_uint)
    {
      g_source_remove (m_repeatTimerId_uint);
      m_repeatTimerId_uint = 0;
    }

  if (g_main_loop_is_running (m_gstMainLoopP))
    g_main_loop_quit (m_gstMainLoopP);

  /* maybe dangerous? */
#ifdef IS_HILDON
  if (m_esd_fd)
    {
      esd_close (m_esd_fd);
      m_esd_fd = 0;
    }
#endif
}
