/**
 * @file gCardSys_Controller.hh
 *
 *  @remark ~ Interface of the gCardSys controller module
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-03 
 **/


#ifndef GCARDSYS_CONTROLLER_HH
#define GCARDSYS_CONTROLLER_HH
/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include "gCardSys_View.hh"
#include "leitnerSysModel.hh"
#include "ViewController_Attrib.hh"

#include <gst/gst.h>            // for gstreamer sound


/*-----------------*
 * Macro Constants *
 \*---------------*/
#define AUTOSOUND_OFF      "0"
#define AUTOSOUND_ON_BACK  "1"
#define AUTOSOUND_ON_FRONT "2"



/*-----------------*
 * Data Structures *
 \*---------------*/
class ControllerClass
{
public:
  ControllerClass (ViewClass      *in_viewObjP,
                   AttributeClass *in_attributeObjP);
  ~ControllerClass ();

  void Init (GtkBuilder *in_builderP,
             const char *in_title_str);
  void CleanUp ();
  void LoadConfig ();

  void SayCurrentWord ();

  ViewClass  *m_viewObjP;  // must be the 1st, see CallBackUserDataStruct
  GtkBuilder *m_builderP;
  ModelClass *m_modelObjP;

  AttributeClass *m_attributeObjP;

  /* sound system -- TBF to be cleanup */
  static gboolean RepeatSayCurrentWord (ControllerClass *in_ctrlObjP);
  void            StopSayCurrentWord ();
  guint           m_repeatTimerId_uint;
  GMainLoop      *m_gstMainLoopP;
  int             m_esd_fd;     // for Maemo n800 device only...

  /* Application state  */
  bool m_isFullScreen_flg;
  bool m_isOpened_flg;
  int  m_frontPageNum_int;
  int  m_backPageNum_int;

private:
  bool PlaySound (const char *in_soundFile_str);

  static gboolean gStreamerBus_call (GstBus     *bus,
                                     GstMessage *msg,
                                     gpointer    data);
  bool m_soundError_flg;
};


/*--------------------*
 * Function Prototype *
 \*------------------*/


#endif // GCARDSYS_CONTROLLER_HH
