/**
 * @file gCardSys_View.cc
 *
 *  @remark ~ Implementation of the gCardSys View module.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright 2009 Tommy Chang.  This file is part of gCardSys.
 *
 *  gCardSys is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  gCardSys is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with gCardSys.  If not, see <http://www.gnu.org/licenses/>.
 *  `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-03 
 **/


/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>             // needed by strcasestr()
#include <unistd.h>             // need by get_current_dir_name

#include "gCardSys_View.hh"


/*---------------------------------------*
 * Global Static Variables and Constants *
 \*-------------------------------------*/
static int fprintf_count = 0;


/*-----------------------------*
 * Functions, macro functions  *
 \*---------------------------*/
#if defined (__MINGW32__) || defined (__GNUC__) 
#  define fprintf(df, ...)                                      \
     do {                                                       \
       if (df == stderr)                                        \
         {                                                      \
           fprintf (stderr, "[gCardSys_View %d]: ",             \
                    fprintf_count++);                           \
           fprintf (stderr, __VA_ARGS__);                       \
           fflush (stderr);                                     \
         }                                                      \
       else                                                     \
         fprintf (df, __VA_ARGS__);                             \
     } while (0)
#endif

#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#ifdef NDEBUG
#define fakeUseParam(...)
#else
static void fakeUseParam(...) {}
#endif


ViewClass::
ViewClass (ModelClass *in_modelObjP)
{
  fprintf (stderr, "calling Constructor\n");

  m_modelObjP = in_modelObjP;
  m_builderP = NULL;
  m_frontText_bufferP = NULL;
  m_backText_bufferP = NULL;
  m_statusText_bufferP = NULL;
}

ViewClass::
~ViewClass () 
{
}


extern "C" G_MODULE_EXPORT void 
on_the_tabs_switch_page (GtkNotebook     *notebook,
                         GtkNotebookPage *page,
                         guint            page_num,
                         gpointer         in_userData)   
{
  fakeUseParam (page);

  CallBackUserDataStruct  *userData_ObjP = 
    (CallBackUserDataStruct*) in_userData;
  ViewClass *view_ObjP = (ViewClass*) userData_ObjP->m_viewObjP;
  GtkBuilder *builderP = view_ObjP->m_builderP;
  
  // fprintf (stderr, "m_builderP is %p\n", builderP);

  GtkWidget *stat_textVP = 
    GTK_WIDGET (gtk_builder_get_object (builderP, "stat_textV"));
  assert (stat_textVP);

  /* Show status page only if requested: */
  guint statPageNum_int = gtk_notebook_page_num (notebook, stat_textVP);
  // fprintf (stderr, "statPageNum_int = %d, page_num = %d\n",
  //          statPageNum_int, page_num);
  if (statPageNum_int == page_num)
    view_ObjP->ConstructStatPage ();
}




bool ViewClass::
InitGUI (GtkBuilder *in_builderP)
{
  fprintf (stderr, "Calling InitGUI\n");

  m_builderP = in_builderP;
  // fprintf (stderr, "m_builderP is %p\n", m_builderP);
  fakeUseParam (in_builderP);

  /*
   * Initialize text views:
   */
  InitTextViews ();


//   GtkWidget *the_tabsP = GTK_WIDGET (gtk_builder_get_object (m_builderP, 
//                                                              "the_tabs"));
//   assert (the_tabsP);
//   g_signal_connect (G_OBJECT (the_tabsP), "key_press_event",
//                     G_CALLBACK (on_the_tabs_key_press_event), this);
  

  return true;
}


void ViewClass::
ConstructStatPage ()
{
  /*
   * Clear the status card:
   */
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (m_statusText_bufferP, &start, &end);
  assert (m_statusText_bufferP);
  gtk_text_buffer_delete (m_statusText_bufferP, &start, &end);

  fprintf (stderr, "prepare status content\n");
  int numCb1DontNeed_int, numCb1Expired_int;
  int numHcbsDontNeed_int, numHcbsExpired_int;
  int numNeverVisited_int;
  int numCardBoxes_int;
  const int *numExpireds_intA;
  if (m_modelObjP->GetNumCards (NULL, &numCb1DontNeed_int, &numCb1Expired_int,
                                &numHcbsDontNeed_int, &numHcbsExpired_int,
                                &numNeverVisited_int, NULL,
                                &numCardBoxes_int, &numExpireds_intA) == false)
    return;


  char status_str[256];
  snprintf (status_str, 256, "there are total %d card\n"
            "%d card in box1 never visited at start of session\n"
            "%d expired cards in box1 now \n"
            "%d expired cards in all higher boxes now\n"
            "\n"
            "Minutes remaining = %d \n\n", 
            numCb1DontNeed_int + numCb1Expired_int +
            numHcbsDontNeed_int + numHcbsExpired_int,
            numNeverVisited_int, numCb1Expired_int, numHcbsExpired_int,
            m_modelObjP->GetRemainStudyTime ());
  
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_offset (m_statusText_bufferP, &iter, 0);
  gtk_text_buffer_insert_with_tags_by_name (m_statusText_bufferP, &iter,
                                            status_str, -1,
                                            "center", NULL);
  /* how many expired cards in each box? :*/
  GString *expiredCard_gstrP = g_string_sized_new (256);
  g_string_printf (expiredCard_gstrP, "Num expired cards in each box: ");
  for (int idx = 0; idx < numCardBoxes_int; idx++)
    g_string_append_printf (expiredCard_gstrP, "%d=%d, ",
                            idx, numExpireds_intA [idx]);
  gtk_text_buffer_insert_with_tags_by_name (m_statusText_bufferP, &iter,
                                            expiredCard_gstrP->str, -1,
                                            "center", NULL);
  g_string_free (expiredCard_gstrP, TRUE);
  
  
  /* Apply word_wrap tag to whole back text buffer */
  gtk_text_buffer_get_bounds (m_statusText_bufferP, &start, &end);
  gtk_text_buffer_apply_tag_by_name (m_statusText_bufferP, 
                                     "word_wrap", &start, &end);
}

static void
ExtractWord (const char *front_str,
             char       *word_str)
{
  /* word may stop at '*':*/
  const char *stop_chrP;
  stop_chrP = strstr (front_str, "*");
  if (stop_chrP)
    sscanf (front_str, "%[^*]", word_str);
  else
    {
      /* word may stop at the category enclosed in '()' */
      stop_chrP = strstr (front_str, "(");
      if (stop_chrP)
        sscanf (front_str, "%[^(]", word_str);
      else
        /* word may stop '\0' */
        sscanf (front_str, "%s", word_str);
    }
  
  /* Remove trailing spaces, in case: */
  for (int idx = strlen (word_str) - 1; idx >= 0; idx --)
    {
      if (word_str [idx] == ' ')
        {
          word_str [idx] = '\0';
          break;
        }
    }
}


void ViewClass::
Present ()
{
  FlashCardInBoxType *flashCardInBoxP = m_modelObjP->GetCurCard ();
  assert (flashCardInBoxP);
  const FlashCardType *flashCardP = flashCardInBoxP->flashCardP;
  assert (flashCardP);
  GtkTextIter iter;
  
  /*
   * The card processed state:
   */
  CardProcessEnum cardProcessed_enum = flashCardInBoxP->processed_enum;
  switch (cardProcessed_enum)
    {
    case CARD_NOT_PROCESSED:    // default setting
    case CARD_DEMOTED:
      {
        GtkToggleButton *nope_radioP = 
          GTK_TOGGLE_BUTTON (gtk_builder_get_object (m_builderP, "nope_radio"));
        assert (nope_radioP);
        m_modelObjP->Demote ();
        fprintf (stderr, "restore  nope setting\n");
        gtk_toggle_button_set_active (nope_radioP, TRUE); // update gui

        // FlashCardInBoxType *flashCardInBoxP = m_modelObjP->GetCurCard ();
        // m_modelObjP->PrintCardInBoxInfo (flashCardInBoxP);
        // fprintf (stderr, "done PrintCardInBoxInfo\n");
      }
      break;
    case CARD_PROMOTED:
    case CARD_MARKED_FOR_PROMOTION:
      {
        GtkToggleButton *yep_radioP = 
          GTK_TOGGLE_BUTTON (gtk_builder_get_object (m_builderP, "yep_radio"));
        assert (yep_radioP);
        m_modelObjP->Promote ();
        fprintf (stderr, "restore  yep setting\n");
        gtk_toggle_button_set_active (yep_radioP, TRUE); // update gui

        // FlashCardInBoxType *flashCardInBoxP = m_modelObjP->GetCurCard ();
        // m_modelObjP->PrintCardInBoxInfo (flashCardInBoxP);
      }
      break;
    case CARD_TRASHED:
      {
        fprintf (stderr, "set forgotten -- not suppose to be here\n");
        abort ();
      }
      break;

    default:
      break;
    }


  /*
   * The front card content:
   */
  gtk_text_buffer_get_iter_at_offset (m_frontText_bufferP, &iter, 0);
  /* have front picture :*/
  if (flashCardP->frontPicFile_str)
    {
      gtk_text_buffer_insert_with_tags_by_name (m_frontText_bufferP, &iter,
                                                flashCardP->front_str, -1,
                                                "center", "big", NULL);
      gtk_text_buffer_insert (m_frontText_bufferP, &iter, "\n", -1);

      /* change current directory: */
      char *curDir_str = get_current_dir_name (); 
      assert (curDir_str);
      // int retVal = chdir (m_modelObjP->GetCardBoxCabinetDir ());
      // assert (retVal == 0);
      int retVal = chdir (flashCardInBoxP->deckObjP->GetFullDeckPath ());
      assert (retVal == 0);

      GdkPixbuf *pixbuf = 
        gdk_pixbuf_new_from_file (flashCardP->frontPicFile_str, NULL);
      
      /* restore the current directory: */
      retVal = chdir (curDir_str);
      free (curDir_str);
      assert (retVal == 0);

      if (pixbuf)
        {
          gtk_text_buffer_insert_pixbuf (m_frontText_bufferP, &iter, pixbuf);
          g_object_unref (pixbuf);
        }
      else
        {
          gchar *missingImg_str = 
            g_strdup_printf ("INVALID image: %s\n",
                             flashCardP->frontPicFile_str);
          gtk_text_buffer_insert (m_frontText_bufferP, &iter, 
                                  missingImg_str, -1);
          g_free (missingImg_str);
        }
    }
  /* No front picture :*/
  else
    {
      gtk_text_buffer_insert (m_frontText_bufferP, &iter, 
                              "\n\n\n\n\n\n", -1);
      gtk_text_buffer_insert_with_tags_by_name (m_frontText_bufferP, &iter,
                                                flashCardP->front_str, -1,
                                                "center", "big", NULL);
    }

  /*
   * The back card content:
   */
  gtk_text_buffer_get_iter_at_offset (m_backText_bufferP, &iter, 0);

  char word_str[80];        // TBF no word is 80 or greater!
  ExtractWord (flashCardP->front_str, word_str);
  fprintf (stderr, "word is '%s'\n", word_str);
    
  int wordLen = strlen (word_str);
  assert (wordLen < 80);
  if (wordLen > 2)                 // TC 2009-10-13 -- was a bug
    word_str [wordLen - 2] = '\0'; // match just abas instead of abase

  /* If in review mode: create sentence-completion exercise: */
  if (m_modelObjP->IsReviewMode ())
    {
      /* mask out word in the definition part: */
      gchar *blanked_str = g_strdup (flashCardP->front_str);
      char *blankedPos_charP = blanked_str;

      if ((blankedPos_charP = strcasestr (blankedPos_charP, word_str)))
        {
          blankedPos_charP [1] = word_str[0]; // give first letter hint
          for (int idx = 1; idx < wordLen; idx++)
            blankedPos_charP [idx] = '_';
        }

      gtk_text_buffer_insert_with_tags_by_name (m_backText_bufferP, &iter,
                                                blanked_str, -1,
                                                "center", NULL);
      g_free (blanked_str);
    }
  /* not review mode: */
  else
    gtk_text_buffer_insert_with_tags_by_name (m_backText_bufferP, &iter,
                                              flashCardP->front_str, -1,
                                              "center", NULL);


  /* the line separator: */
  gtk_text_buffer_insert_with_tags_by_name 
    (m_backText_bufferP, &iter, 
     "\n                                                                \n", -1,
     "center", "red_foreground", "double_underline",  NULL);


  /* the definition: */
  gtk_text_buffer_insert_with_tags_by_name (m_backText_bufferP, &iter,
                                            flashCardP->back_str, -1,
                                            "center", NULL);

  /* the line separator: */
  gtk_text_buffer_insert_with_tags_by_name 
    (m_backText_bufferP, &iter, 
     "\n                                                                \n", -1,
     "center", "red_foreground", "double_underline",  NULL);


  /* the example: */
  /* if under review mode, create sentence-completion exercise: */
  if (strlen (flashCardP->backExample_str) && m_modelObjP->IsReviewMode ())
    {
      /* create sentence-completion exercise: */
      gchar *blankedBackEx_str = g_strdup (flashCardP->backExample_str);
      char searchWord_str[80];        // TBF no word is 80 or greater!
      strncpy (searchWord_str, word_str, 79);

      /* first word of the sentence is the word: */
      fprintf (stderr, "blankedBackEx_str = '%s'\n, searchWord_str = '%s'\n",
               blankedBackEx_str, searchWord_str);
      char *strPos_charP = strcasestr (blankedBackEx_str, searchWord_str);
      if (strPos_charP == blankedBackEx_str) // first word in the example
        {
          strPos_charP [0] = searchWord_str [0]; // give first letter hint
          for (int idx = 1; idx < wordLen; idx++)
            strPos_charP [idx] = '_';
        }

      /* look for rest of sentence for word: */
      strPos_charP = blankedBackEx_str + 1; // don't start from first letter
      while ((strPos_charP = strcasestr (strPos_charP, searchWord_str)))
        {
          /* look for word separation: */
          if (strPos_charP [-1] == ' ' ||
              strPos_charP [-1] == '"' ||
              strPos_charP [-1] == '\'' ||
              strPos_charP [-1] == '-' || // TC 2010-11-10 (Wed) -- was bug
              strPos_charP [-1] == '(')
            {
              strPos_charP [0] = searchWord_str [0]; // give first letter hint
              for (int idx = 1; idx < wordLen; idx++)
                strPos_charP [idx] = '_';
              strPos_charP += strlen (searchWord_str); // search more 
            }
          else
            {
              // TC 2010-11-11 (Thu) -- was bug
              fprintf (stderr, "no word separation for '%s'\n"
                       "in '%s'\n", searchWord_str, blankedBackEx_str);
              strPos_charP ++;  // search more
              break;
            }
        }

      /* set the back card text: */
      gtk_text_buffer_insert (m_backText_bufferP, &iter, blankedBackEx_str, -1);
      g_free (blankedBackEx_str);
    }
  /* not review mode: */
  else
    gtk_text_buffer_insert (m_backText_bufferP, &iter,
                            flashCardP->backExample_str, -1);

  /* have back picture :*/
  if (flashCardP->backPicFile_str)
    {
      gtk_text_buffer_insert (m_backText_bufferP, &iter, "\n", -1);

      /* change current directory: */
      char *curDir_str = get_current_dir_name (); 
      assert (curDir_str);
      // int retVal = chdir (m_modelObjP->GetCardBoxCabinetDir ());
      // assert (retVal == 0);
      int retVal = chdir (flashCardInBoxP->deckObjP->GetFullDeckPath ());
      assert (retVal == 0);
      
      GdkPixbuf *pixbuf = 
        gdk_pixbuf_new_from_file (flashCardP->backPicFile_str, NULL);
      
      /* restore the current directory: */
      retVal = chdir (curDir_str);
      free (curDir_str);
      assert (retVal == 0);

      if (pixbuf)
        {
          gtk_text_buffer_insert_pixbuf (m_backText_bufferP, &iter, pixbuf);
          g_object_unref (pixbuf);
        }
      else
        {
          gchar *missingImg_str = 
            g_strdup_printf ("INVALID image: %s\n",
                             flashCardP->backPicFile_str);
          gtk_text_buffer_insert (m_backText_bufferP, &iter, 
                                  missingImg_str, -1);
          g_free (missingImg_str);
        }
    }


  /* Apply word_wrap tag to whole back text buffer */
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds (m_backText_bufferP, &start, &end);
  gtk_text_buffer_apply_tag_by_name (m_backText_bufferP, 
                                     "word_wrap", &start, &end);

  /*
   * The Status Content (prepare only if needed):
   */
  GtkNotebook *the_tabsP = 
    GTK_NOTEBOOK (gtk_builder_get_object (m_builderP, "the_tabs"));
  assert (the_tabsP);
  GtkWidget *stat_textVP = 
    GTK_WIDGET (gtk_builder_get_object (m_builderP, "stat_textV"));
  assert (stat_textVP);
  gint pageNum_int = gtk_notebook_page_num (the_tabsP, stat_textVP);

  if (gtk_notebook_get_current_page (the_tabsP) == pageNum_int)
    ConstructStatPage ();


  /*
   * Flip to the card if autoFlip is on
   */
  const char *autoFlip_str = m_attributeObj.GetValue ("autoFlip_enum");
  if (strcmp (autoFlip_str, AUTOFLIP_OFF) != 0)
    {
      /* back side under Review mode: */
      if (strcmp (autoFlip_str, AUTOFLIP_TO_BACK) == 0)
        {
          GtkWidget *backCard_scrldWindowP =
            GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                                "backCard_scrldWindow"));
          assert (backCard_scrldWindowP);
          
          pageNum_int = gtk_notebook_page_num (the_tabsP, 
                                               backCard_scrldWindowP);
          assert (pageNum_int > 0);
        }
      else
        /* front side under Study mode: */
        {
          GtkWidget *frontCard_textVP =
            GTK_WIDGET (gtk_builder_get_object (m_builderP, 
                                                "frontCard_textV"));
          assert (frontCard_textVP);
          
          pageNum_int = gtk_notebook_page_num  (the_tabsP, frontCard_textVP);
        }
      gtk_notebook_set_current_page (the_tabsP, pageNum_int);
    }
}


void ViewClass::
InitTextViews ()
{
  GtkTextView *textViewP;

  /*
   * Create some text tags:
   */
  GtkTextTag *bigTextTagP = gtk_text_tag_new ("big");
  g_object_set (bigTextTagP, "size", 20 * PANGO_SCALE, NULL);
  GtkTextTag *centerTextTagP = gtk_text_tag_new ("center");
  g_object_set (centerTextTagP, "justification", GTK_JUSTIFY_CENTER, NULL);
//   GtkTextTag *unEditableTextTagP = gtk_text_tag_new ("not_editable");
//   g_object_set (unEditableTextTagP, "editable", FALSE, NULL);
  GtkTextTag *redTextTagP = gtk_text_tag_new ("red_foreground");
  g_object_set (redTextTagP, "foreground", "red", NULL);
  GtkTextTag *dUnderLineTextTagP = gtk_text_tag_new ("double_underline");
  g_object_set (redTextTagP, "underline", PANGO_UNDERLINE_DOUBLE, NULL);
  GtkTextTag *wrapTextTagP = gtk_text_tag_new ("word_wrap");
  g_object_set (wrapTextTagP, "wrap_mode", GTK_WRAP_WORD, NULL);


  /*
   * Create a shared tag table:
   */
  GtkTextTagTable *gtkTextTableP = gtk_text_tag_table_new ();
  gtk_text_tag_table_add (gtkTextTableP, bigTextTagP);
  gtk_text_tag_table_add (gtkTextTableP, centerTextTagP);
//   gtk_text_tag_table_add (gtkTextTableP, unEditableTextTagP);
  gtk_text_tag_table_add (gtkTextTableP, redTextTagP);
  gtk_text_tag_table_add (gtkTextTableP, dUnderLineTextTagP);
  gtk_text_tag_table_add (gtkTextTableP, wrapTextTagP);


  /*
   * Recreate buffer with new shared tag table for Front of the Card:
   */
  textViewP = GTK_TEXT_VIEW (gtk_builder_get_object (m_builderP, 
                                                     "frontCard_textV"));
  assert (textViewP);
  m_frontText_bufferP = gtk_text_buffer_new (gtkTextTableP); 
  gtk_text_view_set_buffer (textViewP, 
                            m_frontText_bufferP); // donn't need unref
  
  
  /*
   * Recreate buffer with new shared tag table for Back of the Card:
   */
  textViewP = GTK_TEXT_VIEW (gtk_builder_get_object (m_builderP, 
                                                     "backCard_textV"));
  assert (textViewP);
  m_backText_bufferP = gtk_text_buffer_new (gtkTextTableP); 
  gtk_text_view_set_buffer (textViewP,
                            m_backText_bufferP); // donn't need unref

  /*
   * Recreate buffer with new shared tag table for Status:
   */
  textViewP = GTK_TEXT_VIEW (gtk_builder_get_object (m_builderP, 
                                                     "stat_textV"));
  assert (textViewP);
  m_statusText_bufferP = gtk_text_buffer_new (gtkTextTableP); 
  gtk_text_view_set_buffer (textViewP, 
                            m_statusText_bufferP); // donn't need unref
  
}


void ViewClass::
ClearPresentation ()
{
  GtkTextIter start, end;

  /*
   * Clear the front card:
   */
  gtk_text_buffer_get_bounds (m_frontText_bufferP, &start, &end);
  assert (m_frontText_bufferP);
  gtk_text_buffer_delete (m_frontText_bufferP, &start, &end);


  /*
   * Clear the back card:
   */
  gtk_text_buffer_get_bounds (m_backText_bufferP, &start, &end);
  assert (m_backText_bufferP);
  gtk_text_buffer_delete (m_backText_bufferP, &start, &end);


  /*
   * Clear the status card:
   */
  gtk_text_buffer_get_bounds (m_statusText_bufferP, &start, &end);
  assert (m_statusText_bufferP);
  gtk_text_buffer_delete (m_statusText_bufferP, &start, &end);
}
