/**
 * @file flashCard.hh
 *
 *  @remark ~ Structure and interface of flash card.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  Copyright (c) 2009 Tommy Chang. Permission is granted to copy,       .
 *  distribute and/or modify this document under the terms of the GNU    .
 *  Free Documentation License, Version 1.3 or any later version         .
 *  published by the Free Software Foundation; with no Invariant         .
 *  Sections, no Front-Cover Texts, and no Back-Cover Texts. A copy of   .
 *  the license is included in the file fdl-1.3.txt.                     .
 *  `--------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2009-06-07 
 **/


#ifndef FLASHCARD_HH
#define FLASHCARD_HH
/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdint.h>

#include "libxml/xmlreader.h"
#include "dict.h"

/*-----------------*
 * Macro Constants *
 \*---------------*/



/*-----------------*
 * Data Structures *
 \*---------------*/

typedef enum DeckXmlStateEnum {
  DECK_XML_STATE_NONE = 0,
  DECK_XML_STATE_SOUND_PATH,
  DECK_XML_STATE_SOUND_PATH_TEXT,
  DECK_XML_STATE_PICS_PATH,
  DECK_XML_STATE_PICS_PATH_TEXT,
  DECK_XML_STATE_CARD,
  DECK_XML_STATE_FRONT,
  DECK_XML_STATE_FRONT_TEXT,
  DECK_XML_STATE_BACK,
  DECK_XML_STATE_BACK_TEXT,
  DECK_XML_STATE_BACK_EXAMPLE,
} DeckXmlStateEnum;

typedef struct FlashCardType {
  uint32_t     wordUiq_id;
  char        *front_str;
  char        *back_str;
  char        *backExample_str;
  char        *frontPicFile_str;
  char        *backPicFile_str;
} FlashCardType;

class DeckSavingClass
{
public:
  DeckSavingClass (const char *in_deckFile_str,
                   const char *in_soundPath_str);
  DeckSavingClass (const DeckSavingClass &); // avoid compiler warning
  void operator =(const DeckSavingClass &);  // avoid compiler warning
  ~DeckSavingClass ();
  bool InsertCard (const FlashCardType *in_flashCardP);
private:
  void FreeMemory ();
  dict_t *m_dictP;
  FILE   *m_fileP;
  char   *m_soundPath_str;
};

class FlashCardClass
{
public:
  FlashCardClass ();
  FlashCardClass (const FlashCardClass &); // avoid compiler warning
  void operator =(const FlashCardClass &); // avoid compiler warning
  ~FlashCardClass ();

  bool Load (const char *in_deckFile_str,
             int        *out_numCards_intP);
  void Unload ();
  bool isLoaded () {return m_loaded_flg;};

  const FlashCardType** GetAllCards (int *out_numCards_intP);
  const FlashCardType*  GetCard (uint32_t in_card_id);
  const char*           GetSoundPath () {return m_soundPath_str;};
  const char*           GetFullSoundPath () {return m_fullSoundPath_str;};
  const char*           GetFullDeckPath () {return m_fullDeckPath_str;};
  // const char*           GetPicPath () {return m_picPath_str;};

  void PrintCardInfo (const FlashCardType *in_cardP);

  bool ChangeFields (uint32_t     in_wordUiq_id,
                     const char  *in_front_str,
                     const char  *in_back_str,
                     const char  *in_backExample_str,
                     const char  *in_frontPicFile_str,
                     const char  *in_backPicFile_str);
  bool Save (const char *in_saveFile_str);

private:
  void ProcessXmlNode (xmlTextReaderPtr in_readerP);
  static int InsertFlashCardToArray (const void *data, 
                                     const void *user);

  bool             m_loaded_flg;
  DeckXmlStateEnum m_deckXmlState_enum;

  FlashCardType   *m_tmpFlashCardP;
  FlashCardType  **m_flashCardsPA; // simulate serialized

  dict_t          *m_dictP;
  char            *m_soundPath_str;
  char            *m_fullSoundPath_str;
  char            *m_picPath_str;
  char            *m_fullDeckPath_str;
};


/*--------------------*
 * Function Prototype *
 \*------------------*/


#endif // FLASHCARD_HH
