/**
 * @file lookupTable.hh
 *
 *  @remark ~ Interface to the generic look-up table support.
 *
 *  @author ~ Tommy Chang
 *
 *  @verbatim
 *  ,--------------------------------------------------------------------.
 *  . THIS software was developed at the National Institute of           .
 *  . Standards and Technology by employees of the Federal Government    .
 *  . in the course of their official duties. Pursuant to title 17       .
 *  . Section 105 of the United States Code this software is not         .
 *  . subject to copyright protection and is in the public               .
 *  . domain. NIST's [intelligent mobility] software is an experimental  .
 *  . system. NIST assumes no responsibility whatsoever for its use by   .
 *  . other parties, and makes no guarantees, expressed or implied,      .
 *  . about its quality, reliability, or any other characteristic. We    .
 *  . would appreciate acknowledgement if the software is used. This     .
 *  . software can be redistributed and/or modified freely provided      .
 *  . that any derivative works bear some notice that they are derived   .
 *  . from it, and any modified versions bear some notice that they      .
 *  . have been modified.                                                .
 *   `-------------------------------------------------------------------
 *  @endverbatim
 *
 *  @par Created by Tommy Chang, 2010-04-10 (Sat) 
 **/


#ifndef LOOKUPTABLE_HH
#define LOOKUPTABLE_HH
/*----------------------------------*
 * Preprocessor and Include Headers *
 \*--------------------------------*/
#include <stdio.h>
#include <stdint.h>
#include "hash.h"

/*-----------------*
 * Macro Constants *
 \*---------------*/



/*-----------------*
 * Data Structures *
 \*---------------*/
class LookUpTable 
{
public:
  LookUpTable ();
  LookUpTable (const LookUpTable &);    // avoid compiler warning
  void operator =(const LookUpTable &); // avoid compiler warning
  ~LookUpTable ();

  bool SetLookupEntries (int         numKeys,
                         const char *keys_str,
                         const char *vals_str);
  bool SetLookupEntry (const char *key_str,
                       const void *val_str,
                       int         valSize_int);
  const void* GetLookUpEntry (const char *key_str);
  void ResetLookUpTable ();
  void Print (void (*printFun) (const char*, const void*));

private:
  hash_t        *m_table_hashP;
} ;



/*--------------------*
 * Function Prototype *
 \*------------------*/


#endif // LOOKUPTABLE_HH
